/*
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 2001-2002,2004 Yeti (David Necas) <yeti@physics.muni.cz>.
 * (but see below)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
 * This plug-in is just a GIMP interface to the Stereograph renderer by
 * Fabian Januszewski, see renderer.c.
 *
 * Please send plug-in patches or suggestions to me: yeti@physics.muni.cz,
 * and renderer patches and suggestions to Fabian Januszewski too (I have
 * only a vague idea how the whole thing works, and please note I CAN'T SEE
 * STEREOGRAMS! ;-)))
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
   /* To include libgimp/stdplugins-intl.h */
#  define GETTEXT_PACKAGE "gimp20"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include "libgimp/stdplugins-intl.h"

#include "renderer.h"
#include "stereograph.h"

#define PROCEDURE_NAME   "stereograph"
#define DATA_KEY_VALS    "plug_in_stereograph"

enum {
    ENTRY_WIDTH = 75,
    SCALE_WIDTH = 201,

    /* How often progressbar is updated */
    PROGRESS_STEP = 1
};

enum {
    RESPONSE_REVERT = 1,
    RESPONSE_RESET,
    RESPONSE_ABOUT
};

/* These are plug-in settings, converted to renderer settings later */
typedef struct {
    gint     texture;
    gint     antialias;
    gdouble  perspective;
    gdouble  nearest_point;
    gdouble  eye_shift;
    gint     linear_algorithm;
    gint     artefacts_prevent;
    gdouble  x_offset;
    gdouble  y_offset;
} PluginValues;

typedef struct {
    GtkWidget *texture;
    GtkObject *antialias;
    GtkObject *perspective;
    GtkObject *nearest_point;
    GtkObject *eye_shift;
    GtkWidget *linear_algorithm;
    GtkWidget *artefacts_prevent;
    GtkObject *x_offset;
    GtkObject *y_offset;
} PluginControls;

static void     query            (void);
static void     run              (const gchar *name,
                                  gint nparams,
                                  const GimpParam *param,
                                  gint *nreturn_vals,
                                  GimpParam **return_vals);
static gboolean plugin_dialog    (GimpDrawable *drawable);
static void     about_dialog     (GtkWidget *parent);
static void     execute          (GimpDrawable *drawable);
static gboolean texture_constrain(gint32 image_id,
                                  gint32 drawable_id,
                                  gpointer data);
static void     texture_changed  (GtkWidget *combo,
                                  gint *pvals);
static void     refresh_controls (const PluginValues *new_pvals);

static GimpPlugInInfo PLUG_IN_INFO = {
    NULL,  /* init_proc  */
    NULL,  /* quit_proc  */
    query, /* query_proc */
    run,   /* run_proc   */
};

static PluginValues pvals_defaults = {
    -1,    /* texture           */
    4,     /* antialias         */
    5.0,   /* perspective       */
    0.4,   /* nearest_point     */
    0.0,   /* eye_shift         */
    TRUE,  /* linear_algorithm  */
    FALSE, /* artefacts_prevent */
    0.0,   /* x_offset          */
    0.0,   /* y_offset          */
};

static PluginControls pctrl;
static PluginValues pvals; /* = pvals_defaults; */
static PluginValues pvals_old;
struct PARAMS *pParam;  /* renderer params */


MAIN()

static void
query(void)
{
    static GimpParamDef args[] = {
        { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
        { GIMP_PDB_IMAGE, "image", "Input image (unused)" },
        { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
        { GIMP_PDB_DRAWABLE, "texture", "Texture" },
        { GIMP_PDB_INT32, "antialias", "Antialiasing depth" },
        { GIMP_PDB_FLOAT, "perspective", "Distance used for perspective" },
        { GIMP_PDB_FLOAT, "nearest_point", "Depth of nearest point" },
        { GIMP_PDB_FLOAT, "eye_shift", "Observer position horizontal shift" },
        { GIMP_PDB_INT32, "linear_algorithm", "Use linear rendering algorithm" },
        { GIMP_PDB_INT32, "artefacts_prevent", "Try to prevent artefacts" },
        { GIMP_PDB_FLOAT, "x_offset", "Texture relative horizontal offset" },
        { GIMP_PDB_FLOAT, "y_offset", "Texture relative vertical offset" },
    };

    gimp_install_procedure
        (DATA_KEY_VALS,
         "Create a Single Image Stereogram",
         "This plugin creates a Single Image Stereogram "
         "from given base image and a texture.",
         "Fabian Januszewski [renderer] & Yeti [Gimp plug-in]",
         "Fabian Januszewski & Yeti",
         "Sep 2001",
         N_("<Image>/Filters/Map/_Stereograph..."),
         "RGB*, GRAY*",
         GIMP_PLUGIN,
         G_N_ELEMENTS(args), 0,
         args, NULL);
}

static void
run(const gchar *name,
    gint nparams,
    const GimpParam  *param,
    gint *nreturn_vals,
    GimpParam  **return_vals)
{
    static GimpParam values[1];
    GimpDrawable *drawable;
    GimpRunMode run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;

    run_mode = param[0].data.d_int32;

    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
    textdomain(GETTEXT_PACKAGE);

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    drawable = gimp_drawable_get(param[2].data.d_drawable);

    pvals = pvals_defaults;
    switch (run_mode) {
        case GIMP_RUN_INTERACTIVE:
        gimp_get_data(DATA_KEY_VALS, &pvals);

        pvals_old = pvals;

        if (!plugin_dialog(drawable)) {
            gimp_drawable_detach(drawable);
            return;
        }
        break;

        case GIMP_RUN_NONINTERACTIVE:
        if (nparams != 11)
            status = GIMP_PDB_CALLING_ERROR;
        else {
            pvals.texture = (gint)param[3].data.d_int32;
            pvals.antialias = (gint)param[4].data.d_int32;
            pvals.perspective = (gdouble)param[5].data.d_float;
            pvals.nearest_point = (gdouble)param[6].data.d_float;
            pvals.eye_shift = (gdouble)param[7].data.d_float;
            pvals.linear_algorithm = (gint)param[8].data.d_int32;
            pvals.artefacts_prevent = (gint)param[9].data.d_int32;
            pvals.x_offset = (gdouble)param[10].data.d_float;
            pvals.y_offset = (gdouble)param[11].data.d_float;
        }
        break;

        case GIMP_RUN_WITH_LAST_VALS:
        gimp_get_data(DATA_KEY_VALS, &pvals);
        break;

        default:
        break;
    }

    if (status == GIMP_PDB_SUCCESS) {
        if (gimp_drawable_is_rgb(drawable->drawable_id)
            || gimp_drawable_is_gray(drawable->drawable_id)) {
            gimp_progress_init(_("Stereograph..."));

            execute(drawable);

            if (run_mode != GIMP_RUN_NONINTERACTIVE)
                gimp_displays_flush();

            if (run_mode == GIMP_RUN_INTERACTIVE)
                gimp_set_data(DATA_KEY_VALS, &pvals, sizeof(PluginValues));
        }
        else
            status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;
    gimp_drawable_detach(drawable);
}

/* remove some source code redundancies {{{ */
static inline void
yeti_table_attach_to_row(GtkWidget *table,
                         GtkWidget *child,
                         const gint row)
{
    gtk_table_attach(GTK_TABLE(table), child, 0, GTK_TABLE(table)->ncols,
                     row, row+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
}

static inline GtkWidget*
yeti_check_button_new_with_label(const gchar *name,
                                 gint *value,
                                 GCallback cb,
                                 gpointer data)
{
    GtkWidget *check;

    check = gtk_check_button_new_with_mnemonic(name);
    g_signal_connect(check, "toggled",
                     G_CALLBACK(gimp_toggle_button_update), value);
    if (cb)
        g_signal_connect_swapped(check, "toggled", cb, data);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), *value);

    return check;
}

static inline GtkObject*
yeti_scale_entry_new_double(const gchar *name,
                            GtkWidget *table,
                            GtkSizeGroup *sgroup,
                            const gint row,
                            double *value,
                            const double min,
                            const double max,
                            const double step,
                            const double page,
                            const gint digits)
{
    GtkObject *adj;

    *value = CLAMP(*value, min, max);
    adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                               *value, min, max, step, page, digits,
                               TRUE, 0, 0, NULL, NULL);
    g_signal_connect(adj, "value-changed",
                     G_CALLBACK(gimp_double_adjustment_update), value);
    if (sgroup)
        gtk_size_group_add_widget(sgroup, GIMP_SCALE_ENTRY_SPINBUTTON(adj));

    return adj;
}

static inline GtkObject*
yeti_scale_entry_new_int(const gchar *name,
                         GtkWidget *table,
                         GtkSizeGroup *sgroup,
                         const gint row,
                         gint *value,
                         const gint min,
                         const gint max)
{
    GtkObject *adj;

    *value = CLAMP(*value, min, max);
    adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                               *value, min, max, 1, 10, 0,
                               TRUE, 0, 0, NULL, NULL);
    g_signal_connect(G_OBJECT(adj), "value_changed",
                     G_CALLBACK(gimp_int_adjustment_update),
                     value);
    if (sgroup)
        gtk_size_group_add_widget(sgroup, GIMP_SCALE_ENTRY_SPINBUTTON(adj));

    return adj;
}

void
yeti_message_dialog(GtkMessageType type,
                    GtkWindow *parent,
                    gchar *title,
                    gchar *message,
                    gchar *secondary)
{
    GtkWidget *dialog;

    dialog = gtk_message_dialog_new(parent, GTK_DIALOG_MODAL,
                                    type, GTK_BUTTONS_OK, message);
    if (title)
        gtk_window_set_title(GTK_WINDOW(dialog), title);
    if (secondary)
        gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
                                                 secondary);
    gtk_window_set_wmclass(GTK_WINDOW(dialog), "message", "Gimp");
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}
/* }}} */

static gboolean
plugin_dialog(GimpDrawable *drawable)
{
    GtkWidget *dialog, *label, *table;
    GtkSizeGroup *sgroup;
    gint row, run;
    gboolean ok;

    gimp_ui_init(PROCEDURE_NAME, TRUE);

    dialog = gimp_dialog_new(_("Stereograph"), PROCEDURE_NAME,
                             NULL, 0,
                             gimp_standard_help_func,
                             "filters/stereograph.html",

                             _("About"), RESPONSE_ABOUT,
                             _("Re_vert"), RESPONSE_REVERT,
                             GIMP_STOCK_RESET, RESPONSE_RESET,
                             GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                             GTK_STOCK_OK, GTK_RESPONSE_OK,
                             NULL);

    sgroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

    table = gtk_table_new(3, 4, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 10);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), table,
                       TRUE, TRUE, 0);
    row = 0;

    /* texture */
    label = gtk_label_new(NULL);
    gtk_label_set_markup(GTK_LABEL(label), _("<b>Texture</b>"));
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    yeti_table_attach_to_row(table, label, row);
    row++;

    if (!texture_constrain(0, pvals.texture, drawable))
        pvals.texture = -1;
    pctrl.texture = gimp_drawable_combo_box_new(texture_constrain, drawable);
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.texture),
                                  pvals.texture);
    gimp_int_combo_box_get_active(GIMP_INT_COMBO_BOX(pctrl.texture),
                                  &pvals.texture);
    g_signal_connect(pctrl.texture, "changed",
                     G_CALLBACK(texture_changed), &pvals.texture);
    gimp_table_attach_aligned(GTK_TABLE(table), 0, row, _("_Texture:"),
                              0.0, 0.5, pctrl.texture, 3, TRUE);
    row++;

    pctrl.x_offset = yeti_scale_entry_new_double(_("Relative _x-offset:"),
                                                 table, sgroup, row,
                                                 &pvals.x_offset,
                                                 0.0, 1.0, 0.01, 0.1, 2);
    row++;

    pctrl.y_offset = yeti_scale_entry_new_double(_("Relative _y-offset:"),
                                                 table, sgroup, row,
                                                 &pvals.y_offset,
                                                 0.0, 1.0, 0.01, 0.1, 2);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    /* renderer options */
    label = gtk_label_new(NULL);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_label_set_markup(GTK_LABEL(label), _("<b>Renderer settings</b>"));
    yeti_table_attach_to_row(table, label, row);
    row++;

    pctrl.antialias = yeti_scale_entry_new_int(_("_Antialiasing depth:"),
                                               table, sgroup, row,
                                               &pvals.antialias,
                                               1, 32);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    pctrl.perspective = yeti_scale_entry_new_double(_("_Perspective distance:"),
                                                    table, sgroup, row,
                                                    &pvals.perspective,
                                                    0.01, 20.0, 0.01, 0.5, 2);
    row++;

    pctrl.nearest_point = yeti_scale_entry_new_double(_("_Nearest point:"),
                                                      table, sgroup, row,
                                                      &pvals.nearest_point,
                                                      0.01, 1.0, 0.01, 0.1, 2);
    row++;

    pctrl.eye_shift = yeti_scale_entry_new_double(_("_Eye shift:"),
                                                  table, sgroup, row,
                                                  &pvals.eye_shift,
                                                  -0.5, 0.5, 0.01, 0.1, 2);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    pctrl.linear_algorithm
        = yeti_check_button_new_with_label(_("_Linear rendering algorithm"),
                                           &pvals.linear_algorithm,
                                           NULL, NULL);
    yeti_table_attach_to_row(table, pctrl.linear_algorithm, row);
    row++;

    pctrl.artefacts_prevent
        = yeti_check_button_new_with_label(_("Reduce arti_facts"),
                                           &pvals.artefacts_prevent,
                                           NULL, NULL);
    yeti_table_attach_to_row(table, pctrl.artefacts_prevent, row);
    row++;

    g_object_unref(sgroup);
    gtk_widget_show_all(dialog);

    ok = FALSE;
    do {
        run = gimp_dialog_run(GIMP_DIALOG(dialog));
        switch (run) {
            case RESPONSE_REVERT:
            refresh_controls(&pvals_old);
            break;

            case RESPONSE_RESET:
            refresh_controls(&pvals_defaults);
            break;

            case RESPONSE_ABOUT:
            about_dialog(dialog);
            break;

            case GTK_RESPONSE_OK:
            if (pvals.texture != -1) {
                ok = TRUE;
                break;
            }
            yeti_message_dialog(GTK_MESSAGE_ERROR, GTK_WINDOW(dialog),
                                _("No Texture"),
                                _("No texture is selected."),
                                _("If you cannot select a texture, it means "
                                  "no drawable is narrow enough to be usable "
                                  "as a texture for this base image."));
            break;

            default:
            ok = TRUE;
            break;
        }
    } while (!ok);

    gtk_widget_destroy(dialog);

    return run == GTK_RESPONSE_OK;
}

static void
about_dialog(GtkWidget *parent)
{
    yeti_message_dialog(GTK_MESSAGE_INFO, GTK_WINDOW(parent),
                        _("About"),
                        _("Stereograph GIMP plug-in " VERSION),
                        _("Based on Stereograph " STEREOGRAPH_VERSION "\n"
                          "http://stereograph.sourceforge.net/\n\n"
                          "Stereograph by Fabian Januszewski.\n"
                          "GIMP plug-in by Yeti."));
}

/* a reasonable contraints are:
   texture must not be wider than half of base (this also cleverly disables
   offering the base drawable itself as a texture)
   it must be either grayscale or RGB */
static gboolean
texture_constrain(gint32 image_id, gint32 drawable_id, gpointer data)
{
    GimpDrawable *drawable;

    if (drawable_id == -1)
        return TRUE;
    drawable = (GimpDrawable*)data;
    if ((gimp_drawable_is_rgb(drawable_id)
         || gimp_drawable_is_gray(drawable_id))
        && 2*gimp_drawable_width(drawable_id) <= drawable->width)
        return TRUE;
    else
        return FALSE;
}

static void
texture_changed(GtkWidget *combo,
                gint *pval)
{
    gimp_int_combo_box_get_active(GIMP_INT_COMBO_BOX(combo), pval);
}

/* though this is obscure, we use it to reset/revert all values _and_ refresh
   the dialog controls */
static void
refresh_controls(const PluginValues *new_pvals)
{
    /* we cannot reset some field and don't want to reset some others, so copy
       them one by one  */
    pvals.linear_algorithm = new_pvals->linear_algorithm;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.linear_algorithm),
                                 pvals.linear_algorithm);

    pvals.artefacts_prevent = new_pvals->artefacts_prevent;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.artefacts_prevent),
                                 pvals.artefacts_prevent);

    pvals.x_offset = new_pvals->x_offset;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.x_offset),
                             pvals.x_offset);

    pvals.y_offset = new_pvals->y_offset;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.y_offset),
                             pvals.y_offset);

    pvals.antialias = new_pvals->antialias;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.antialias),
                             pvals.antialias);

    pvals.perspective = new_pvals->perspective;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.perspective),
                             pvals.perspective);

    pvals.nearest_point = new_pvals->nearest_point;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.nearest_point),
                             pvals.nearest_point);

    pvals.eye_shift = new_pvals->eye_shift;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.eye_shift),
                             pvals.eye_shift);

    /* don't want to update: texture */
}

/****************************************************************************/

static void
execute(GimpDrawable *drawable)
{
    gint32 stereo_lay, stereo_img;
    GimpDrawable *texture, *stereo;
    struct GFX_DATA base_gfx, texture_gfx, stereo_gfx;
    struct RENDERER_PARAMS rparam;
    struct RENDERER_DATA rdata;
    GimpPixelRgn base_region, texture_region, stereo_region;
    guchar *base_row, *texture_row, *stereo_row; /* Gimp row buffers */
    gint base_c, texture_c, stereo_c, base_bpp, texture_bpp, stereo_bpp;
    gint y, x, b, c;

    /* this is only renderer's way of making aliases and thus is safe to call
       anytime */
    Renderer_GFX_Init(&rdata, &base_gfx, &texture_gfx, &stereo_gfx);

    /* initialize base and texture (always allocate one line only) */
    texture_gfx.Width = gimp_drawable_width(pvals.texture);
    texture_gfx.Height = gimp_drawable_height(pvals.texture);
    texture_gfx.Data = g_new(gint32, texture_gfx.Width);
    texture = gimp_drawable_get(pvals.texture);

    base_gfx.Width = gimp_drawable_width(drawable->drawable_id);
    base_gfx.Height = gimp_drawable_height(drawable->drawable_id);
    base_gfx.Data = g_new(gint32, base_gfx.Width);

    /* No.  We should definitely always supply our own values. */
    /* Renderer_Param_Init(&rparam); */

    /* copy plug-in parameters to renderer parameters */
    rparam.AA = pvals.antialias;
    rparam.Zoom = 1; /* not useful at all in GIMP? */
    rparam.Distance = pvals.perspective;
    rparam.Front = pvals.nearest_point;
    rparam.Eyeshift = pvals.eye_shift;
    rparam.T_Layers = 0; /* FIXME: hard to implement */
    rparam.Linear = pvals.linear_algorithm;
    rparam.AAr = pvals.artefacts_prevent;
    rparam.StartX = pvals.x_offset*texture_gfx.Width;
    rparam.StartY = pvals.y_offset*texture_gfx.Height;

    /* XXX: renderer must be patched not to allocate Stereo */
    x = Renderer_Initialize(&rdata, &rparam);
    g_assert(x == 0); /* FIXME: do gods still like me? */

    /* Renderer_Initialize() computes Width & Height so we can allocate
     * memory */
    stereo_gfx.Data = g_new(gint32, stereo_gfx.Width);

    /* create a new image */
    stereo_img = gimp_image_new(stereo_gfx.Width, stereo_gfx.Height, GIMP_RGB);
    gimp_image_undo_disable(stereo_img);
    stereo_lay = gimp_layer_new(stereo_img, "Stereograph",
                                stereo_gfx.Width, stereo_gfx.Height,
                                GIMP_RGB_IMAGE, 100, GIMP_NORMAL_MODE);
    gimp_image_add_layer(stereo_img, stereo_lay, 0);
    stereo = gimp_drawable_get(stereo_lay);

    /* set up all the tile and pixel region stuff */
    gimp_tile_cache_ntiles((stereo_gfx.Width + base_gfx.Width
                            + texture_gfx.Width)/gimp_tile_width() + 2);
    gimp_pixel_rgn_init(&base_region, drawable, 0, 0,
                        base_gfx.Width, base_gfx.Height, FALSE, FALSE);
    gimp_pixel_rgn_init(&texture_region, texture, 0, 0,
                        texture_gfx.Width, texture_gfx.Height, FALSE, FALSE);
    gimp_pixel_rgn_init(&stereo_region, stereo, 0, 0,
                        stereo_gfx.Width, stereo_gfx.Height, TRUE, TRUE);

    /* allocate row buffers */
    base_c = base_bpp = gimp_drawable_bpp(drawable->drawable_id);
    if (gimp_drawable_has_alpha(drawable->drawable_id))
        base_c--;
    base_row = g_new(guchar, base_bpp*base_gfx.Width);

    texture_c = texture_bpp = gimp_drawable_bpp(pvals.texture);
    if (gimp_drawable_has_alpha(pvals.texture))
        texture_c--;
    texture_row = g_new(guchar, texture_bpp*texture_gfx.Width);

    stereo_c = stereo_bpp = gimp_drawable_bpp(stereo->drawable_id);
    if (gimp_drawable_has_alpha(stereo->drawable_id))
        stereo_c--;
    stereo_row = g_new(guchar, stereo_bpp*stereo_gfx.Width);

    /* render it! */
    for (y = 0; y < base_gfx.Height; y++) {
        /* read base row */
        gimp_pixel_rgn_get_row(&base_region, base_row, 0, y, base_gfx.Width);
        memset(base_gfx.Data, 0, sizeof(gint32)*base_gfx.Width);
        for (c = b = 0; b < 3; b++) {
            for (x = 0; x < base_gfx.Width; x++)
                base_gfx.Data[x] |= (gint32)(base_row[x*base_bpp + c]) << 8*b;
            if (c < base_c - 1)
                c++;
        }

        /* read texture row */
        gimp_pixel_rgn_get_row(&texture_region, texture_row, 0,
                               (y + rparam.StartY)%texture_gfx.Height,
                               texture_gfx.Width);
        memset(texture_gfx.Data, 0, sizeof(gint32)*texture_gfx.Width);
        for (c = b = 0; b < 3; b++) {
            for (x = 0; x < texture_gfx.Width; x++) {
                texture_gfx.Data[x]
                    |= (gint32)(texture_row[x*texture_bpp + c]) << 8*b;
            }
            if (c < texture_c - 1)
                c++;
        }

        /* call renderer to do its work */
        Renderer_RenderLine(base_gfx.Data, texture_gfx.Data, stereo_gfx.Data,
                            y, &rdata);

        /* set alpha to maximum and move result to the image */
        memset(stereo_row, 0xff, stereo_bpp*stereo_gfx.Width);
        for (b = 0; b < 3; b++) {
            for (x = 0; x < stereo_gfx.Width; x++) {
                stereo_row[x*stereo_bpp + b]
                    = (stereo_gfx.Data[x] >> 8*b) & 0xff;
            }
        }
        gimp_pixel_rgn_set_row(&stereo_region,
                               stereo_row, 0, y, stereo_gfx.Width);
        if (y % PROGRESS_STEP == PROGRESS_STEP - 1)
            gimp_progress_update((gdouble)y/stereo_gfx.Height);
    }

    /* clean-up */
    Renderer_Close(&rdata);

    g_free(base_row);
    g_free(texture_row);
    g_free(stereo_row);
    g_free(base_gfx.Data);
    g_free(texture_gfx.Data);
    g_free(stereo_gfx.Data);

    gimp_drawable_flush(stereo);
    gimp_drawable_merge_shadow(stereo->drawable_id, TRUE);
    gimp_drawable_update(stereo->drawable_id,
                         0, 0, stereo_gfx.Width, stereo_gfx.Height);
    gimp_drawable_detach(stereo);

    gimp_image_undo_enable(stereo_img);
    gimp_image_clean_all(stereo_img);
    gimp_display_new(stereo_img);
}

